class Path:
    __slots__ = ("method", "path", "scope", "requires_user")

    def __init__(self, method, path, scope, requires_user=False):
        self.method = method
        self.path = path
        self.scope = scope
        self.requires_user = requires_user

    @property
    def requires_auth(self):
        return self.scope is not None

    @classmethod
    def beatmap_lookup(cls):
        return cls('get', "beatmaps/lookup", 'public')

    @classmethod
    def user_beatmap_score(cls, beatmap, user):
        return cls('get', f"beatmaps/{beatmap}/scores/users/{user}", 'public')

    @classmethod
    def user_beatmap_scores(cls, beatmap, user):
        return cls('get', f"beatmaps/{beatmap}/scores/users/{user}/all", 'public')

    @classmethod
    def beatmap_scores(cls, beatmap):
        return cls('get', f"beatmaps/{beatmap}/scores", 'public')

    @classmethod
    def lazer_beatmap_scores(cls, beatmap):
        return cls('get', f"beatmaps/{beatmap}/solo-scores", 'public')

    @classmethod
    def beatmap(cls, beatmap):
        return cls('get', f"beatmaps/{beatmap}", 'public')

    @classmethod
    def beatmaps(cls):
        return cls('get', "beatmaps", 'public')

    @classmethod
    def get_beatmap_attributes(cls, beatmap):
        return cls('post', f"beatmaps/{beatmap}/attributes", 'public')

    @classmethod
    def get_beatmapset(cls, beatmapset):
        return cls('get', f'beatmapsets/{beatmapset}', 'public')

    @classmethod
    def beatmapset_discussion_posts(cls):
        return cls('get', 'beatmapsets/discussions/posts', 'public')

    @classmethod
    def beatmapset_discussion_votes(cls):
        return cls('get', 'beatmapsets/discussions/votes', 'public')

    @classmethod
    def beatmapset_discussions(cls):
        return cls('get', 'beatmapsets/discussions', 'public')

    @classmethod
    def get_changelog_build(cls, stream, build):
        return cls('get', f"changelog/{stream}/{build}", None)

    @classmethod
    def get_changelog_listing(cls):
        return cls('get', 'changelog', None)

    @classmethod
    def lookup_changelog_build(cls, changelog):
        return cls('get', f'changelog/{changelog}', None)

    @classmethod
    def chat_ack(cls):
        return cls('post', 'chat/ack', 'lazer', True)

    @classmethod
    def create_new_pm(cls):
        return cls('post', 'chat/new', 'chat.write', True)

    @classmethod
    def get_updates(cls):
        return cls('get', 'chat/updates', 'lazer', True)

    @classmethod
    def get_channel_messages(cls, channel):
        return cls('get', f'chat/channels/{channel}/messages', 'lazer', True)

    @classmethod
    def send_message_to_channel(cls, channel):
        return cls('post', f'chat/channels/{channel}/messages', 'lazer', True)

    @classmethod
    def join_channel(cls, channel, user):
        return cls('put', f'chat/channels/{channel}/users/{user}', 'lazer', True)

    @classmethod
    def leave_channel(cls, channel, user):
        return cls('delete', f'chat/channels/{channel}/users/{user}', 'lazer', True)

    @classmethod
    def mark_channel_as_read(cls, channel, message):
        return cls('put', f'chat/channels/{channel}/mark-as-read/{message}', 'lazer', True)

    @classmethod
    def get_channel_list(cls):
        return cls('get', 'chat/channels', 'lazer', True)

    @classmethod
    def create_channel(cls):
        return cls('post', 'chat/channels', 'lazer', True)

    @classmethod
    def get_channel(cls, channel):
        return cls('get', f'chat/channels/{channel}', 'lazer', True)

    @classmethod
    def get_comments(cls):
        return cls('get', 'comments', None)

    @classmethod
    def post_new_comment(cls):
        return cls('post', 'comments', 'lazer', True)

    @classmethod
    def get_comment(cls, comment):
        return cls('get', f'comments/{comment}', None)

    @classmethod
    def edit_comment(cls, comment):
        return cls('patch', f'comments/{comment}', 'lazer', True)

    @classmethod
    def delete_comment(cls, comment):
        return cls('delete', f'comments/{comment}', 'lazer', True)

    @classmethod
    def add_comment_vote(cls, comment):
        return cls('post', f'comments/{comment}/vote', 'lazer', True)

    @classmethod
    def remove_comment_vote(cls, comment):
        return cls('delete', f'comments/{comment}/vote', 'lazer', True)

    @classmethod
    def reply_topic(cls, topic):
        return cls('post', f'forums/topics/{topic}/reply', 'forum.write', True)

    @classmethod
    def create_topic(cls):
        return cls('post', 'forums/topics', 'forum.write', True)

    @classmethod
    def get_topic_and_posts(cls, topic):
        return cls('get', f'forums/topics/{topic}', 'public')

    @classmethod
    def edit_topic(cls, topic):
        return cls('patch', f'forums/topics/{topic}', 'forum.write')

    @classmethod
    def edit_post(cls, post):
        return cls('patch', f'forums/posts/{post}', 'forum.write')

    @classmethod
    def search(cls):
        return cls('get', 'search', 'public')

    @classmethod
    def get_user_high_score(cls, room, playlist, user):
        return cls('get', f'rooms/{room}/playlist/{playlist}/scores/users/{user}', 'lazer', True)

    @classmethod
    def get_scores(cls, room, playlist):
        return cls('get', f'rooms/{room}/playlist/{playlist}/scores', 'public', True)

    @classmethod
    def get_score(cls, room, playlist, score):
        return cls('get', f'rooms/{room}/playlist/{playlist}/scores/{score}', 'lazer', True)

    @classmethod
    def get_news_listing(cls):
        return cls('get', 'news', None)

    @classmethod
    def get_news_post(cls, news):
        return cls('get', f'news/{news}', None)

    @classmethod
    def get_notifications(cls):
        return cls('get', 'notifications', 'lazer', True)

    @classmethod
    def mark_notifications_as_read(cls):
        return cls('post', 'notifications/mark-read', 'lazer', True)

    @classmethod
    def revoke_current_token(cls):
        return cls('delete', 'oauth/tokens/current', 'public')

    @classmethod
    def get_ranking(cls, mode, type):
        return cls('get', f'rankings/{mode}/{type}', 'public')

    @classmethod
    def get_spotlights(cls):
        return cls('get', 'spotlights', 'public')

    @classmethod
    def get_own_data(cls, mode=''):
        return cls('get', f'me/{mode}', 'identify')

    @classmethod
    def get_user_kudosu(cls, user):
        return cls('get', f'users/{user}/kudosu', 'public')

    @classmethod
    def get_user_scores(cls, user, type):
        return cls('get', f'users/{user}/scores/{type}', 'public')

    @classmethod
    def get_user_beatmaps(cls, user, type):
        return cls('get', f'users/{user}/beatmapsets/{type}', 'public')

    @classmethod
    def get_user_recent_activity(cls, user):
        return cls('get', f'users/{user}/recent_activity', 'public')

    @classmethod
    def get_user(cls, user, mode=''):
        return cls('get', f'users/{user}/{mode}', 'public')

    @classmethod
    def get_users(cls):
        return cls('get', 'users', 'public')

    @classmethod
    def get_wiki_page(cls, locale, path):
        return cls('get', f'wiki/{locale}/{path}', None)

    @classmethod
    def get_score_by_id(cls, mode, score):
        return cls('get', f'scores/{mode}/{score}', 'public')

    @classmethod
    def get_beatmapset_events(cls):
        return cls('get', 'beatmapsets/events', 'public')

    @classmethod
    def get_matches(cls):
        return cls('get', 'matches', 'public')

    @classmethod
    def get_match(cls, match):
        return cls('get', f'matches/{match}', 'public')

    @classmethod
    def get_rooms(cls, mode=''):
        return cls('get', f'rooms/{mode}', 'public', True)

    @classmethod
    def get_room(cls, room):
        return cls('get', f'rooms/{room}', 'public')

    @classmethod
    def get_room_leaderboard(cls, room):
        return cls('get', f'rooms/{room}/leaderboard', 'public', True)

    @classmethod
    def get_seasonal_backgrounds(cls):
        return cls('get', 'seasonal-backgrounds', None)

    @classmethod
    def get_replay_data(cls, mode, score):
        return cls('get', f'scores/{mode}/{score}/download', 'public', True)

    @classmethod
    def get_friends(cls):
        return cls('get', 'friends', 'friends.read', True)

    @classmethod
    def get_new_score_id(cls, beatmap_id):
        return cls('post', f'beatmaps/{beatmap_id}/solo/scores', 'lazer', True)

    @classmethod
    def submit_score(cls, beatmap_id, token):
        return cls('put', f'beatmaps/{beatmap_id}/solo/scores/{token}', 'lazer', True)

    @classmethod
    def favourite_beatmapset(cls, beatmapset_id):
        return cls('post', f'beatmapsets/{beatmapset_id}/favourites', 'lazer', True)

    @classmethod
    def get_chat_presence(cls):
        return cls('get', 'chat/presence', 'lazer', True)

    @classmethod
    def join_to_room(cls, room, user):
        return cls('put', f'rooms/{room}/users/{user}', 'lazer', True)

    @classmethod
    def kick_from_room(cls, room, user):
        return cls('delete', f'rooms/{room}/users/{user}', 'lazer', True)

    @classmethod
    def send_report(cls):
        return cls('post', 'reports', 'lazer', True)

    @classmethod
    def create_room(cls):
        return cls('post', 'rooms', 'lazer', True)

    @classmethod
    def download_quota_check(cls):
        return cls('get', 'me/download-quota-check', 'lazer', True)

    @classmethod
    def beatmapset_search(cls):
        return cls('get', 'beatmapsets/search', 'public')



from enum import IntFlag, IntEnum, Enum
from typing import Sequence, Union
from .constants import incompatible_mods


class Mod(Enum):
    """
    Enum of all mods, score submittable or not.

    **Mods**

    Easy = 'EZ'

    NoFail = 'NF'

    HalfTime = 'HT'

    Daycore = 'DC'

    HardRock = 'HR'

    SuddenDeath = 'SD'

    Perfect = 'PF'

    DoubleTime = 'DT'

    Nightcore = 'NC'

    Hidden = 'HD'

    Flashlight = 'FL'

    Blinds = 'BL'

    StrictTracking = 'ST'

    Target = 'TP'

    DifficultyAdjust = 'DA'

    Classic = 'CL'

    Random = 'RD'

    Mirror = 'MR'

    Alternate = 'AL'

    SingleTap = 'SG'

    Autoplay = 'AT'

    Cinema = 'CN'

    Relax = 'RX'

    Autopilot = 'AP'

    SpunOut = 'SO'

    Transform = 'TR'

    Wiggle = 'WG'

    SpinIn = 'SI'

    Grow = 'GR'

    Deflate = 'DF'

    WindUp = 'WU'

    WindDown = 'WD'

    Traceable = 'TC'

    BarrelRoll = 'BR'

    ApproachDifferent = 'AD'

    Muted = 'MU'

    NoScope = 'NS'

    Magnetised = 'MG'

    Repel = 'RP'

    AdaptiveSpeed = 'AS'

    FreezeFrame = 'FR'

    TouchDevice = 'TD'

    Swap = 'SW'

    FloatingFruits = 'FF'

    FadeIn = 'FI'

    FourKeys = '4K'

    FiveKeys = '5K'

    SixKeys = '6K'

    SevenKeys = '7K'

    EightKeys = '8K'

    NineKeys = '9K'

    TenKeys = '10K'

    OneKey = '1K'

    TwoKeys = '2K'

    ThreeKeys = '3K'

    DualStages = 'DS'

    Invert = 'IN'

    ConstantSpeed = 'CS'

    HoldOff = 'HO'

    AccuracyChallenge = 'AC'
    """
    Easy = 'EZ'
    NoFail = 'NF'
    HalfTime = 'HT'
    Daycore = 'DC'
    HardRock = 'HR'
    SuddenDeath = 'SD'
    Perfect = 'PF'
    DoubleTime = 'DT'
    Nightcore = 'NC'
    Hidden = 'HD'
    Flashlight = 'FL'
    Blinds = 'BL'
    StrictTracking = 'ST'
    Target = 'TP'
    DifficultyAdjust = 'DA'
    Classic = 'CL'
    Random = 'RD'
    Mirror = 'MR'
    Alternate = 'AL'
    SingleTap = 'SG'
    Autoplay = 'AT'
    Cinema = 'CN'
    Relax = 'RX'
    Autopilot = 'AP'
    SpunOut = 'SO'
    Transform = 'TR'
    Wiggle = 'WG'
    SpinIn = 'SI'
    Grow = 'GR'
    Deflate = 'DF'
    WindUp = 'WU'
    WindDown = 'WD'
    Traceable = 'TC'
    BarrelRoll = 'BR'
    ApproachDifferent = 'AD'
    Muted = 'MU'
    NoScope = 'NS'
    Magnetised = 'MG'
    Repel = 'RP'
    AdaptiveSpeed = 'AS'
    FreezeFrame = 'FR'
    TouchDevice = 'TD'
    Swap = 'SW'
    FloatingFruits = 'FF'
    FadeIn = 'FI'
    FourKeys = '4K'
    FiveKeys = '5K'
    SixKeys = '6K'
    SevenKeys = '7K'
    EightKeys = '8K'
    NineKeys = '9K'
    TenKeys = '10K'
    OneKey = '1K'
    TwoKeys = '2K'
    ThreeKeys = '3K'
    DualStages = 'DS'
    Invert = 'IN'
    ConstantSpeed = 'CS'
    HoldOff = 'HO'
    AccuracyChallenge = 'AC'


class Mods(IntFlag):
    """
    IntFlag enum for all score submittable mods. Info gathered from
    https://github.com/ppy/osu-web/blob/973315aded8a5762fc00a9f245337802c27bd213/app/Libraries/Mods.php
    and https://github.com/ppy/osu-web/blob/973315aded8a5762fc00a9f245337802c27bd213/database/mods.json

    **List of mods, their acronyms, and their bitwise representation**

    NoFail (NF) = 1 << 0

    Easy (EZ) = 1 << 1

    TouchDevice (TD) = 1 << 2  # Replaces unused NoVideo mod

    Hidden (HD) = 1 << 3

    HardRock (HR) = 1 << 4

    SuddenDeath (SD) = 1 << 5

    DoubleTime (DT) = 1 << 6

    Relax (RX) = 1 << 7

    HalfTime (HT) = 1 << 8

    Nightcore (NC) = (1 << 9)

    Flashlight (FL) = 1 << 10

    SpunOut (SO) = 1 << 12

    AutoPilot (AP) = 1 << 13

    Perfect (PF) = 1 << 14

    FadeIn (FI) = 1 << 20

    Mirror (MR) = 1 << 30

    Key4 (4K) = 1 << 15

    Key5 (5K) = 1 << 16

    Key6 (6K) = 1 << 17

    Key7 (7K) = 1 << 18

    Key8 (8K) = 1 << 19

    Key9 (9K) = 1 << 24
    """

    NoFail = 1 << 0
    Easy = 1 << 1
    TouchDevice = 1 << 2
    Hidden = 1 << 3
    HardRock = 1 << 4
    SuddenDeath = 1 << 5
    DoubleTime = 1 << 6
    Relax = 1 << 7
    HalfTime = 1 << 8
    Nightcore = (1 << 9)
    Flashlight = 1 << 10
    SpunOut = 1 << 12
    AutoPilot = 1 << 13
    Perfect = 1 << 14
    FadeIn = 1 << 20
    Mirror = 1 << 30

    FourKeys = 1 << 15
    FiveKeys = 1 << 16
    SixKeys = 1 << 17
    SevenKeys = 1 << 18
    EightKeys = 1 << 19
    NineKeys = 1 << 24

    @classmethod
    def get_from_abbreviation(cls, abbreviation: str) -> 'Mods':
        """
        Get mod from its abbreviation. Abbreviations are taken from https://osu.ppy.sh/wiki/en/Game_modifier/Summary

        **Parameters**

        abbreviation: :class:`str`
            Abbreviation of the mod (must be capitalized)

        **Returns**

        :class:`Mods`
        """
        return cls[Mod(abbreviation.upper()).name]

    @staticmethod
    def get_from_list(mods: Sequence['Mods']) -> Union['Mods', None]:
        """
        Get a :class:`Mods` object from a list of :class:`Mods`.

        **Parameters**

        mods: Sequence[:class:`Mods`]
            Sequence of mods of type Mods

        **Returns**

        Union[:class:`Mods`, :class:`NoneType`]
        """
        if len(mods) == 0:
            return
        a = mods[0]
        for i in range(1, len(mods)):
            a |= mods[i]
        return a

    @staticmethod
    def parse_any_list(mods: Sequence[Union[str, int, 'Mods']]) -> 'Mods':
        """
        Take a list and return a parsed list. Parsing the list involves
        converting any object recognizable as a mod to a :class:`Mods` object.
        This includes mod names/abbreviations as strings and also their bitset values.

        **Parameters**

        mods: Sequence[Union[:class:`Mods`, :class:`str`, :class:`int`]]
            Sequence of :class:`Mods`, :class:`str`, and/or :class:`int` objects
            to be parsed and returned as a :class:`Mods` object.

        **Returns**

        Union[:class:`Mods`, :class:`NoneType`]
        """
        ret = []
        for mod in mods:
            if isinstance(mod, Mods):
                ret.append(mod)
            elif isinstance(mod, str):
                if mod in Mods.__members__:
                    ret.append(Mods[mod])
                else:
                    try:
                        ret.append(Mods.get_from_abbreviation(mod.upper()))
                    except ValueError:
                        raise ValueError("Mods represented as strings must be either the full name or "
                                         "abbreviation. '{mod}' does not fall under either of those.")
            elif isinstance(mod, int):
                ret.append(Mods(mod))
            else:
                raise TypeError("Mods can only be parsed to Mods objects if they're of type str, int, or Mods")
        return Mods.get_from_list(ret) if len(ret) > 0 else None

    def get_incompatible_mods(self):
        """
        Get a list of mods that are incompatible with this mod.

        **Returns**

        Sequence[:class:`Mods`]
        """
        if self.name is None:
            raise ValueError("Cannot get incompatible mods of a multi-mods enum object.")
        return list(map(lambda x: Mods[x], incompatible_mods[self.name]))

    def is_compatible_with(self, other: 'Mods'):
        """
        Check if this mod is compatible with another mod.

        **Parameters**

        other: :class:`Mods`
            Mod to check compatibility with.

        **Returns**

        :class:`bool`
        """
        if self.name is None or other.name is None:
            raise ValueError("Cannot check compatibility of a multi-mods enum object.")
        return other not in self.get_incompatible_mods()

    def is_compatible_combination(self):
        """
        Check if all the mods in this Mods object are compatible with each other.

        **Returns**

        :class:`bool`
        """
        mods = list(self)

        for i in range(len(mods)):
            for j in range(i+1, len(mods)):
                if not mods[i].is_compatible_with(mods[j]):
                    return False
        return True

    def to_readable_string(self):
        """
        Get a readable string representation of this mod (sorted by bitset ascending).
        Example: (Mods.HardRock | Mods.Hidden) -> "HDHR"

        **Returns**

        :class:`str`
        """
        return "".join(map(lambda mod: Mod[mod.name].value,
                           sorted(self, key=lambda m: m.value)))

    def __iter__(self):
        mods = str(self).split('|')
        mods[0] = mods[0].replace("Mods.", "")
        mods = list(map(lambda x: Mods[x], mods))
        return iter(mods)


class RankStatus(IntEnum):
    """
    IntEnum enum for rank status of a beatmap.

    **Statuses**

    GRAVEYARD = -2

    WIP = -1

    PENDING = 0

    RANKED = 1

    APPROVED = 2

    QUALIFIED = 3

    LOVED = 4
    """

    GRAVEYARD = -2
    WIP = -1
    PENDING = 0
    RANKED = 1
    APPROVED = 2
    QUALIFIED = 3
    LOVED = 4


class GameModeStr(Enum):
    """
    Enum for GameModes using their string names.

    **GameModes**

    STANDARD = 'osu'

    TAIKO = 'taiko'

    CATCH = 'fruits'

    MANIA = 'mania'
    """
    STANDARD = 'osu'
    TAIKO = 'taiko'
    CATCH = 'fruits'
    MANIA = 'mania'

    @staticmethod
    def get_int_equivalent(gamemode: Enum) -> IntEnum:
        return GameModeInt[gamemode.name]


class GameModeInt(IntEnum):
    """
    Enums for GameModes using their int values.

    **GameModes**

    STANDARD = 0

    TAIKO = 1

    CATCH = 2

    MANIA = 3
    """
    STANDARD = 0
    TAIKO = 1
    CATCH = 2
    MANIA = 3

    @staticmethod
    def get_str_equivalent(gamemode: IntEnum) -> Enum:
        return GameModeStr[gamemode.name]


class WikiSearchMode(Enum):
    """
    Enum for wiki search modes. Relevant to :func:`osu.Client.search`.
    """
    ALL = 'all'
    USER = 'user'
    WIKI = 'wiki_page'


class UserBeatmapType(Enum):
    """
    User beatmap types. Relavent to :func:`osu.Client.get_user_beatmaps`.

    **User beatmap types**

    favourite, graveyard, loved, most_played, pending, ranked

    FAVOURITE = 'favourite'

    GRAVEYARD = 'graveyard'

    LOVED = 'loved'

    MOST_PLAYED = 'most_played'

    PENDING = 'pending'

    RANKED = 'ranked'
    """

    FAVOURITE = 'favourite'
    GRAVEYARD = 'graveyard'
    LOVED = 'loved'
    MOST_PLAYED = 'most_played'
    PENDING = 'pending'
    RANKED = 'ranked'


class RankingType(Enum):
    """
    Ranking types to sort by for :func:`osu.Client.get_ranking`.

    **Ranking types**

    SPOTLIGHT = 'charts'

    COUNTRY = 'country'

    PERFORMANCE = 'performance'

    SCORE = 'score'
    """
    SPOTLIGHT = 'charts'
    COUNTRY = 'country'
    PERFORMANCE = 'performance'
    SCORE = 'score'


class CommentSort(Enum):
    """
    Type to sort comments by. Relevant to :func:`osu.Client.get_comments`.

    **Comment sorts**

    NEW = 'new'

    OLD = 'old'

    TOP = 'top'
    """
    NEW = 'new'
    OLD = 'old'
    TOP = 'top'


class MultiplayerScoresSort(Enum):
    """
    Sort option for multiplayer scores index. Relevant to :func:`osu.Client.get_scores`.

    **Multiplayer scores sorts**

    ASC = 'score_asc'

    DESC = 'score_desc'
    """
    ASC = 'score_asc'
    DESC = 'score_desc'


class BeatmapsetEventType(Enum):
    """
    Enum for beatmapset event types. Relevant to :func:`osu.Client.get_beatmapset_events`.

    **Beatmapset event types**

    NOMINATE = 'nominate'

    LOVE = 'love'

    REMOVE_FROM_LOVED = 'unlove'

    QUALIFY = 'qualify'

    DISQUALIFY = 'disqualify'

    APPROVE = 'approve'

    RANK = 'rank'

    KUDOSU_ALLOW = 'kudosu_allow'

    KUDOSU_DENY = 'kudosu_deny'

    KUDOSU_GAIN = 'kudosu_gain'

    KUDOSU_LOST = 'kudosu_lost'

    KUDOSU_RECALCULATE = 'kudosu_recalculate'

    ISSUE_RESOLVE = 'issue_resolve'

    ISSUE_REOPEN = 'issue_reopen'

    DISCUSSION_LOCK = 'discussion_lock'

    DISCUSSION_UNLOCK = 'discussion_unlock'

    DISCUSSION_DELETE = 'discussion_delete'

    DISCUSSION_RESTORE = 'discussion_restore'

    DISCUSSION_POST_DELETE = 'discussion_post_delete'

    DISCUSSION_POST_RESTORE = 'discussion_post_restore'

    NOMINATION_RESET = 'nomination_reset'

    NOMINATION_RESET_RECEIVED = 'nomination_reset_received'

    GENRE_EDIT = 'genre_edit'

    LANGUAGE_EDIT = 'language_edit'

    NSFW_TOGGLE = 'nsfw_toggle'

    OFFSET_EDIT = 'offset_edit'

    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'
    """

    NOMINATE = 'nominate'
    LOVE = 'love'
    REMOVE_FROM_LOVED = 'remove_from_loved'
    QUALIFY = 'qualify'
    DISQUALIFY = 'disqualify'
    APPROVE = 'approve'
    RANK = 'rank'

    KUDOSU_ALLOW = 'kudosu_allow'
    KUDOSU_DENY = 'kudosu_deny'
    KUDOSU_GAIN = 'kudosu_gain'
    KUDOSU_LOST = 'kudosu_lost'
    KUDOSU_RECALCULATE = 'kudosu_recalculate'

    ISSUE_RESOLVE = 'issue_resolve'
    ISSUE_REOPEN = 'issue_reopen'

    DISCUSSION_LOCK = 'discussion_lock'
    DISCUSSION_UNLOCK = 'discussion_unlock'

    DISCUSSION_DELETE = 'discussion_delete'
    DISCUSSION_RESTORE = 'discussion_restore'

    DISCUSSION_POST_DELETE = 'discussion_post_delete'
    DISCUSSION_POST_RESTORE = 'discussion_post_restore'

    NOMINATION_RESET = 'nomination_reset'
    NOMINATION_RESET_RECEIVED = 'nomination_reset_received'

    GENRE_EDIT = 'genre_edit'
    LANGUAGE_EDIT = 'language_edit'
    NSFW_TOGGLE = 'nsfw_toggle'
    OFFSET_EDIT = 'offset_edit'

    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'


class BeatmapsetEventSort(Enum):
    """
    Sort option for beatmapset events. Relevant to :func:`osu.Client.get_beatmapset_events`.

    **Beatmapset event sorts**

    ASC = 'id_asc'

    DESC = 'id_desc'
    """
    ASC = 'id_asc'
    DESC = 'id_desc'


class MatchSort(Enum):
    """
    Sort options for matches. Relevant to :func:`osu.Client.get_matches`.

    **Match sort options**

    ID_ASC = 'id_asc'

    ID_DESC = 'id_desc'

    NAME_ASC = 'name_asc'

    NAME_DESC = 'name_desc'
    """
    OLDEST = 'id_asc'
    NEWEST = 'id_desc'


class MatchEventType(Enum):
    """
    Enum for match event types.

    **Match event types**

    PLAYER_LEFT = 'player-left'

    PLAYER_JOINED = 'player-joined'

    PLAYER_KICKED = 'player-kicked'

    MATCH_CREATED = 'match-created'

    MATCH_DISBANDED = 'match-disbanded'

    HOST_CHANGED = 'host-changed'

    OTHER = 'other'
    """
    PLAYER_LEFT = 'player-left'
    PLAYER_JOINED = 'player-joined'
    PLAYER_KICKED = 'player-kicked'
    MATCH_CREATED = 'match-created'
    MATCH_DISBANDED = 'match-disbanded'
    HOST_CHANGED = 'host-changed'
    OTHER = 'other'


class RoomSort(Enum):
    """
    Sort options for rooms. Relevant to :func:`osu.Client.get_rooms`.

    **Room sort options**

    ENDED = 'ended'

    CREATED = 'created'
    """
    ENDED = 'ended'
    CREATED = 'created'


class RoomCategory(Enum):
    """
    Enum for room categories.

    **Room categories**

    NORMAL = 'normal'

    SPOTLIGHT = 'spotlight'

    FEATURED_ARTIST = 'featured_artist'
    """
    NORMAL = 'normal'
    SPOTLIGHT = 'spotlight'
    FEATURED_ARTIST = 'featured_artist'


class RoomType(Enum):
    """
    Enum for room types.

    **Room types**

    PLAYLISTS = 'playlists'

    REALTIME = 'realtime'
    """
    PLAYLISTS = 'playlists'
    REALTIME = 'realtime'


class RoomFilterMode(Enum):
    """
    Enum for different filtering modes of rooms.

    **Room filter mode types**

    ENDED = 'ended'

    PARTICIPATED = 'participated'

    OWNED = 'owned'

    ACTIVE = 'active'
    """
    ENDED = 'ended'
    PARTICIPATED = 'participated'
    OWNED = 'owned'
    ACTIVE = 'active'


class RealTimeType(Enum):
    """
    Enum for realtime types.

    **Realtime types**

    HEAD_TO_HEAD = 'head_to_head'

    TEAM_VERSUS = 'team_versus'
    """
    HEAD_TO_HEAD = 'head_to_head'
    TEAM_VERSUS = 'team_versus'


class RealTimeQueueMode(Enum):
    """
    Enum for realtime queue modes.

    **Realtime queue modes**

    HOST_ONLY = 'host_only'

    ALL_PLAYERS = 'all_players'

    ALL_PLAYERS_ROUND_ROBIN = 'all_players_round_robin'
    """
    HOST_ONLY = 'host_only'
    ALL_PLAYERS = 'all_players'
    ALL_PLAYERS_ROUND_ROBIN = 'all_players_round_robin'


class PlaylistQueueMode(Enum):
    """
    Enum for playlist queue modes.

    **Playlist queue modes**

    HOST_ONLY = 'host_only'
    """
    HOST_ONLY = 'host_only'


class BeatmapsetSearchSort(Enum):
    """
    Sort options for beatmapset searches. Relevant to :func:`osu.Client.search_beatmapsets`.

    **Beatmapset search sort options**

    ARTIST = 'artist'

    CREATOR = 'creator'

    DIFFICULTY = 'difficulty'

    FAVOURITES = 'favourites'

    NOMINATIONS = 'nominations'

    PLAYS = 'plays'

    RANKED = 'ranked'

    RATING = 'rating'

    RELEVANCE = 'relevance'

    TITLE = 'title'

    UPDATED = 'updated'
    """
    ARTIST = 'artist'
    CREATOR = 'creator'
    DIFFICULTY = 'difficulty'
    FAVOURITES = 'favourites'
    NOMINATIONS = 'nominations'
    PLAYS = 'plays'
    RANKED = 'ranked'
    RATING = 'rating'
    RELEVANCE = 'relevance'
    TITLE = 'title'
    UPDATED = 'updated'


class BeatmapsetSearchStatus(Enum):
    """
    Status options for beatmapset filtering. Relevant to :func:`osu.Client.search_beatmapsets`.

    **Beatmapset search status options**

    ANY = 'any'

    HAS_LEADERBOARD = 'leaderboard'

    RANKED = 'ranked'

    QUALIFIED = 'qualified'

    LOVED = 'loved'

    FAVOURITES = 'favourites'

    PENDING = 'pending'

    WIP = 'wip'

    GRAVEYARD = 'graveyard'

    MY_MAPS = 'mine'
    """
    ANY = 'any'
    HAS_LEADERBOARD = 'leaderboard'
    RANKED = 'ranked'
    QUALIFIED = 'qualified'
    LOVED = 'loved'
    FAVOURITES = 'favourites'
    PENDING = 'pending'
    WIP = 'wip'
    GRAVEYARD = 'graveyard'
    MY_MAPS = 'mine'


class BeatmapsetSearchExtra(Enum):
    """
    Extra options for beatmapset filtering. Relevant to :func:`osu.Client.search_beatmapsets`.

    **Beatmapset search extra options**

    VIDEO = 'video'

    STORYBOARD = 'storyboard'
    """
    VIDEO = 'video'
    STORYBOARD = 'storyboard'


class BeatmapsetSearchGeneral(Enum):
    """
    General options for beatmapset filtering. Relevant to :func:`osu.Client.search_beatmapsets`.

    **Beatmapset search general options**

    RECOMMENDED = 'recommended'

    CONVERTS = 'converts'

    FOLLOWS = 'follows'

    SPOTLIGHTS = 'spotlights'

    FEATURED_ARTISTS = 'featured_artists'
    """
    RECOMMENDED = 'recommended'
    CONVERTS = 'converts'
    FOLLOWS = 'follows'
    SPOTLIGHTS = 'spotlights'
    FEATURED_ARTISTS = 'featured_artists'


class BeatmapsetSearchPlayed(Enum):
    """
    Played options for beatmapset filtering. Relevant to :func:`osu.Client.search_beatmapsets`.

    **Beatmapset search played options**

    ANY = 'any'

    PLAYED = 'played'

    UNPLAYED = 'unplayed'
    """
    ANY = 'any'
    PLAYED = 'played'
    UNPLAYED = 'unplayed'


class ScoreRank(Enum):
    """
    Enum for score ranks.

    **Score ranks**

    SILVER_SS = 'XH'

    SS = 'X'

    SILVER_S = 'SH'

    S = 'S'

    A = 'A'

    B = 'B'

    C = 'C'

    D = 'D'
    """
    SILVER_SS = 'XH'
    SS = 'X'
    SILVER_S = 'SH'
    S = 'S'
    A = 'A'
    B = 'B'
    C = 'C'
    D = 'D'


class ObjectType(Enum):
    """
    Enum for different object types. Score types are most relevant.

    **Object types**

    BeatmapDiscussion = "beatmapset_discussion"

    BeatmapDiscussionPost = "beatmapset_discussion_post"

    Beatmapset = "beatmapset"

    Build = "build"

    Channel = "channel"

    Comment = "comment"

    ForumPost = "forum_post"

    ForumTopic = "forum_topic"

    LegacyMatchScore = "legacy_match_score"

    NewsPost = "news_post"

    ScoreBestFruits = "score_best_fruits"

    ScoreBestMania = "score_best_mania"

    ScoreBestOsu = "score_best_osu"

    ScoreBestTaiko = "score_best_taiko"

    ScoreFruits = "score_fruits"

    ScoreMania = "score_mania"

    ScoreOsu = "score_osu"

    ScoreTaiko = "score_taiko"

    SoloScore = "solo_score"

    User = "user"
    """
    BeatmapDiscussion = "beatmapset_discussion"
    BeatmapDiscussionPost = "beatmapset_discussion_post"
    Beatmapset = "beatmapset"
    Build = "build"
    Channel = "channel"
    Comment = "comment"
    ForumPost = "forum_post"
    ForumTopic = "forum_topic"
    LegacyMatchScore = "legacy_match_score"
    NewsPost = "news_post"
    ScoreBestFruits = "score_best_fruits"
    ScoreBestMania = "score_best_mania"
    ScoreBestOsu = "score_best_osu"
    ScoreBestTaiko = "score_best_taiko"
    ScoreFruits = "score_fruits"
    ScoreMania = "score_mania"
    ScoreOsu = "score_osu"
    ScoreTaiko = "score_taiko"
    SoloScore = "solo_score"
    User = "user"


class UserScoreType(Enum):
    """
    Enum for the get_user_scores endpoint that specifies score type.

    **User score types**

    BEST = "best"

    FIRSTS = "firsts"

    RECENT = "recent"
    """
    BEST = "best"
    FIRSTS = "firsts"
    RECENT = "recent"


class ForumTopicType(Enum):
    """
    Enum for :class:`ForumTopic` type attribute.

    **Forum topic types**

    NORMAL = "normal"

    STICKY = "sticky"

    ANNOUNCEMENT = "announcement"
    """
    NORMAL = "normal"
    STICKY = "sticky"
    ANNOUNCEMENT = "announcement"


class ChatChannelType(Enum):
    """
    Enum for type of :class:`ChatChannel`

    **Chat channel types**

    PUBLIC = "PUBLIC"

    PRIVATE = "PRIVATE"

    MULTIPLAYER = "MULTIPLAYER"

    SPECTATOR = "SPECTATOR"

    TEMPORARY = "TEMPORARY"

    PM = "PM"

    GROUP = "GROUP"

    ANNOUNCE = "ANNOUNCE"
    """
    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    MULTIPLAYER = "MULTIPLAYER"
    SPECTATOR = "SPECTATOR"
    TEMPORARY = "TEMPORARY"
    PM = "PM"
    GROUP = "GROUP"
    ANNOUNCE = "ANNOUNCE"


class NotificationType(Enum):
    """
    Types of notifications that can be received.

    **Notification types**

    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'

    BEATMAPSET_DISCUSSION_LOCK = 'beatmapset_discussion_lock'

    BEATMAPSET_DISCUSSION_POST_NEW = 'beatmapset_discussion_post_new'

    BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM = 'beatmapset_discussion_qualified_problem'

    BEATMAPSET_DISCUSSION_REVIEW_NEW = 'beatmapset_discussion_review_new'

    BEATMAPSET_DISCUSSION_UNLOCK = 'beatmapset_discussion_unlock'

    BEATMAPSET_DISQUALIFY = 'beatmapset_disqualify'

    BEATMAPSET_LOVE = 'beatmapset_love'

    BEATMAPSET_NOMINATE = 'beatmapset_nominate'

    BEATMAPSET_QUALIFY = 'beatmapset_qualify'

    BEATMAPSET_RANK = 'beatmapset_rank'

    BEATMAPSET_REMOVE_FROM_LOVED = 'beatmapset_remove_from_loved'

    BEATMAPSET_RESET_NOMINATIONS = 'beatmapset_reset_nominations'

    CHANNEL_ANNOUNCEMENT = 'channel_announcement'

    CHANNEL_MESSAGE = 'channel_message'

    COMMENT_NEW = 'comment_new'

    FORUM_TOPIC_REPLY = 'forum_topic_reply'

    USER_ACHIEVEMENT_UNLOCK = 'user_achievement_unlock'

    USER_BEATMAPSET_NEW = 'user_beatmapset_new'

    USER_BEATMAPSET_REVIVE = 'user_beatmapset_revive'
    """
    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'
    BEATMAPSET_DISCUSSION_LOCK = 'beatmapset_discussion_lock'
    BEATMAPSET_DISCUSSION_POST_NEW = 'beatmapset_discussion_post_new'
    BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM = 'beatmapset_discussion_qualified_problem'
    BEATMAPSET_DISCUSSION_REVIEW_NEW = 'beatmapset_discussion_review_new'
    BEATMAPSET_DISCUSSION_UNLOCK = 'beatmapset_discussion_unlock'
    BEATMAPSET_DISQUALIFY = 'beatmapset_disqualify'
    BEATMAPSET_LOVE = 'beatmapset_love'
    BEATMAPSET_NOMINATE = 'beatmapset_nominate'
    BEATMAPSET_QUALIFY = 'beatmapset_qualify'
    BEATMAPSET_RANK = 'beatmapset_rank'
    BEATMAPSET_REMOVE_FROM_LOVED = 'beatmapset_remove_from_loved'
    BEATMAPSET_RESET_NOMINATIONS = 'beatmapset_reset_nominations'
    CHANNEL_ANNOUNCEMENT = 'channel_announcement'
    CHANNEL_MESSAGE = 'channel_message'
    COMMENT_NEW = 'comment_new'
    FORUM_TOPIC_REPLY = 'forum_topic_reply'
    USER_ACHIEVEMENT_UNLOCK = 'user_achievement_unlock'
    USER_BEATMAPSET_NEW = 'user_beatmapset_new'
    USER_BEATMAPSET_REVIVE = 'user_beatmapset_revive'

    def get_category(self):
        return NotificationCategory[self.name]


class NotificationCategory(Enum):
    """
    Enum for notification categories.

    **Attributes**

    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'

    BEATMAPSET_DISCUSSION_LOCK = 'beatmapset_discussion'

    BEATMAPSET_DISCUSSION_POST_NEW = 'beatmapset_discussion'

    BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM = 'beatmapset_problem'

    BEATMAPSET_DISCUSSION_REVIEW_NEW = 'beatmapset_discussion'

    BEATMAPSET_DISCUSSION_UNLOCK = 'beatmapset_discussion'

    BEATMAPSET_DISQUALIFY = 'beatmapset_state'

    BEATMAPSET_LOVE = 'beatmapset_state'

    BEATMAPSET_NOMINATE = 'beatmapset_state'

    BEATMAPSET_QUALIFY = 'beatmapset_state'

    BEATMAPSET_RANK = 'beatmapset_state'

    BEATMAPSET_REMOVE_FROM_LOVED = 'beatmapset_state'

    BEATMAPSET_RESET_NOMINATIONS = 'beatmapset_state'

    CHANNEL_ANNOUNCEMENT = 'announcement'

    CHANNEL_MESSAGE = 'channel'

    COMMENT_NEW = 'comment'

    FORUM_TOPIC_REPLY = 'forum_topic_reply'

    USER_ACHIEVEMENT_UNLOCK = 'user_achievement_unlock'

    USER_BEATMAPSET_NEW = 'user_beatmapset_new'

    USER_BEATMAPSET_REVIVE = 'user_beatmapset_new'
    """
    BEATMAP_OWNER_CHANGE = 'beatmap_owner_change'
    BEATMAPSET_DISCUSSION_LOCK = 'beatmapset_discussion'
    BEATMAPSET_DISCUSSION_POST_NEW = 'beatmapset_discussion'
    BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM = 'beatmapset_problem'
    BEATMAPSET_DISCUSSION_REVIEW_NEW = 'beatmapset_discussion'
    BEATMAPSET_DISCUSSION_UNLOCK = 'beatmapset_discussion'
    BEATMAPSET_DISQUALIFY = 'beatmapset_state'
    BEATMAPSET_LOVE = 'beatmapset_state'
    BEATMAPSET_NOMINATE = 'beatmapset_state'
    BEATMAPSET_QUALIFY = 'beatmapset_state'
    BEATMAPSET_RANK = 'beatmapset_state'
    BEATMAPSET_REMOVE_FROM_LOVED = 'beatmapset_state'
    BEATMAPSET_RESET_NOMINATIONS = 'beatmapset_state'
    CHANNEL_ANNOUNCEMENT = 'announcement'
    CHANNEL_MESSAGE = 'channel'
    COMMENT_NEW = 'comment'
    FORUM_TOPIC_REPLY = 'forum_topic_reply'
    USER_ACHIEVEMENT_UNLOCK = 'user_achievement_unlock'
    USER_BEATMAPSET_NEW = 'user_beatmapset_new'
    USER_BEATMAPSET_REVIVE = 'user_beatmapset_new'



from ..util import prettify


class WikiPage:
    """
    Represents a wiki article

    **Attributes**

    available_locales: Sequence[:class:`str`]
        All available locales for the article.

    layout: :class:`str`
        The layout type for the page.

    locale: :class:`str`
        All lowercase BCP 47 language tag.

    markdown: :class:`str`
        Markdown content.

    path: :class:`str`
        Path of the article.

    subtitle: :class:`str` or :class:`NoneType`
        The article's subtitle.

    tags: Sequence[:class:`str`]
        Associated tags for the article.

    title: :class:`str`
        The article's title.
    """
    __slots__ = (
        'available_locales', 'layout', 'locale', 'markdown', 'path', 'subtitle',
        'tags', 'title'
    )

    def __init__(self, data):
        self.available_locales = data['available_locales']
        self.layout = data['layout']
        self.locale = data['locale']
        self.markdown = data['markdown']
        self.path = data['path']
        self.subtitle = data['subtitle']
        self.tags = data['tags']
        self.title = data['title']

    def __repr__(self):
        return prettify(self, 'title')


class SearchResults:
    """
    Represents the results of a search.

    **Attributes**

    results: Union[:class:`UserCompact`, :class:`WikiPage`]
        type depends on search type

    total: :class:`int`
    """
    __slots__ = ("results", "total")

    def __init__(self, data, data_type):
        self.results = list(map(data_type, data["data"]))
        self.total = data["total"]

    def __repr__(self):
        return prettify(self, 'results', 'total')



from ..enums import RankStatus, GameModeStr, GameModeInt
from .user import UserCompact, CurrentUserAttributes
from dateutil import parser

from ..util import prettify


class BeatmapsetCompact:
    """
    Represents a beatmapset.

    **Attributes**

    artist: :class:`str`

    artist_unicode: :class:`str`

    covers: :class:`Covers`

    creator: :class:`str`

    favourite_count: :class:`int`

    id: :class:`int`

    nsfw: :class:`bool`

    play_count: :class:`int`

    preview_url: :class:`str`

    source: :class:`str`

    status: :class:`RankStatus`

    title: :class:`str`

    title_unicode: :class:`str`

    user_id: :class:`int`

    video: :class:`str`

    **Possible Attributes**

    (Information about these attributes is lacking on the documentation)

    beatmaps: :class:`list`
        list containing objects of type :class:`Beatmap`

    converts

    current_user_attributes: :class:`CurrentUserAttributes`

    description

    discussions

    events

    genre

    has_favourited: :class:`bool`

    language

    nominations: :class:`dict`
        Contains items current: :class:`int` and required: :class:`int`

    ratings

    recent_favourites

    related_users

    user: :class:`UserCompact`
    """
    __slots__ = (
        "artist", "artist_unicode", "covers", "creator", "favourite_count", "id", "nsfw",
        "play_count", "preview_url", "source", "status", "title", "title_unicode", "user_id",
        "video", "beatmaps", "current_user_attributes", "user", "converts", "description", "discussions",
        "events", "genre", "has_favourited", "language", "nominations", "ratings", "recent_favourites",
        "related_users"
    )

    def __init__(self, data):
        self.artist = data['artist']
        self.artist_unicode = data['artist_unicode']
        self.covers = Covers(data['covers'])
        self.creator = data['creator']
        self.favourite_count = data['favourite_count']
        self.id = data['id']
        self.nsfw = data['nsfw']
        self.play_count = data['play_count']
        self.preview_url = data['preview_url']
        self.source = data['source']
        self.status = RankStatus[data['status'].upper()]
        self.title = data['title']
        self.title_unicode = data['title_unicode']
        self.user_id = data['user_id']
        self.video = data['video']

        # Documentation lacks information on all the possible attributes :/
        self.beatmaps = list(map(Beatmap, data['beatmaps'])) if 'beatmaps' in data else None
        self.current_user_attributes = CurrentUserAttributes(data['current_user_attributes'],
                                                             'BeatmapsetDiscussionPermissions') \
            if 'current_user_attributes' in data else None
        self.user = UserCompact(data['user']) if 'user' in data else None
        self.converts = data.get('converts')
        self.description = data.get('description')
        self.discussions = data.get('discussions')
        self.events = data.get('events')
        self.genre = data.get('genre')
        self.has_favourited = data.get('has_favourited')
        self.language = data.get('language')
        self.nominations = data.get('nominations')
        self.ratings = data.get('ratings')
        self.recent_favourites = data.get('recent_favourites')
        self.related_users = data.get('related_users')

    def __repr__(self):
        return prettify(self, 'artist', 'title', 'creator')


class Covers:
    """
    **Attributes**

    cover: :class:`str`

    cover_2x: :class:`str`

    card: :class:`str`

    card_2x: :class:`str`

    list: :class:`str`

    list_2x: :class:`str`

    slimcover: :class:`str`

    slimcover_2x: :class:`str`
    """
    __slots__ = (
        "cover", "cover_2x", "card", "card_2x", "list", "list_2x", "slimcover", "slimcover_2x"
    )

    def __init__(self, data):
        self.cover = data['cover']
        self.cover_2x = data['cover@2x']
        self.card = data['card']
        self.card_2x = data['card@2x']
        self.list = data['list']
        self.list_2x = data['list@2x']
        self.slimcover = data['slimcover']
        self.slimcover_2x = data['slimcover@2x']

    def __repr__(self):
        return prettify(self, 'cover')


class Beatmapset(BeatmapsetCompact):
    """
    Represents a beatmapset. This extends :class:`BeatmapsetCompact` with additional attributes.

    **Attributes**

    availability: :class:`dict`
        Contains two items, download_disabled: :class:`bool` and more_information: :class:`str`

    bpm: :class:`float`

    can_be_hyped: :class:`bool`

    creator: :class:`str`
        Username of the mapper at the time of beatmapset creation.

    discussion_enabled: :class:`bool`
        Deprecated. Is always true.

    discussion_locked: :class:`bool`

    hype: :class:`dict`
        Contains items current: :class:`int` and required: :class:`int`

    is_scoreable: :class:`bool`

    last_updated: :class:`datetime.datetime`

    legacy_thread_url: :class:`str`

    nominations: :class:`dict`
        Contains items current: :class:`int` and required: :class:`int`

    ranked: :class:`RankStatus`

    ranked_date: :class:`datetime.datetime`

    source: :class:`str`

    storyboard: :class:`bool`

    submitted_date: :class:`datetime.datetime`

    tags: :class:`str`
    """
    __slots__ = (
        "availability", "bpm", "can_be_hyped", "creator", "discussion_enabled", "discussion_locked",
        "hype", "is_scoreable", "last_updated", "legacy_thread_url", "ranked", "ranked_date", "storyboard",
        "submitted_date", "tags", "has_favourited", "nominations"
    )

    def __init__(self, data):
        super().__init__(data)
        self.availability = data['availability']
        self.bpm = data['bpm']
        self.can_be_hyped = data['can_be_hyped']
        self.creator = data['creator']
        self.discussion_enabled = True  # Deprecated, all beatmapset discussions are enabled
        self.discussion_locked = data['discussion_locked']
        self.hype = data['hype']
        self.is_scoreable = data['is_scoreable']
        self.last_updated = parser.parse(data['last_updated']) if data.get('last_updated') is not None else None
        self.legacy_thread_url = data['legacy_thread_url']
        self.ranked_date = parser.parse(data['ranked_date']) if data.get('ranked_date') is not None else None
        self.source = data['source']
        self.storyboard = data['storyboard']
        self.tags = data['tags']
        self.submitted_date = parser.parse(data['submitted_date']) if data.get('submitted_date') is not None else None
        self.has_favourited = data['has_favourited'] if 'has_favourited' in data else None
        self.ranked = RankStatus(int(data['ranked']))
        self.nominations = data['nominations'] if 'nominations' in data else None

    def __repr__(self):
        return super().__repr__()


class BeatmapCompact:
    """
    Represents a beatmap.

    **Attributes**

    beatmapset_id: :class:`int`

    difficulty_rating: :class:`float`

    id: :class:`int`

    mode: :class:`GameModeStr`

    status: :class:`RankStatus`

    total_length: :class:`int`

    user_id: :class:`int`

    version: :class:`str`

    **Possible Attributes**

    beatmapset: :class:`Beatmapset` | :class:`BeatmapsetCompact` | :class:`NoneType`
        Beatmapset for Beatmap object, BeatmapsetCompact for BeatmapCompact object.
        null if the beatmap doesn't have associated beatmapset (e.g. deleted).

    checksum: :class:`str` or :class:`NoneType`

    failtimes: :class:`Failtimes`

    max_combo: :class:`int`
    """
    __slots__ = (
        "difficulty_rating", "id", "mode", "status", "total_length", "user_id", "version",
        "checksum", "max_combo", "failtimes", "beatmapset"
    )

    def __init__(self, data):
        self.difficulty_rating = data['difficulty_rating']
        self.id = data['id']
        self.mode = GameModeStr(data['mode'])
        self.status = RankStatus[data['status'].upper()]
        self.total_length = data['total_length']
        self.user_id = data['user_id']
        self.version = data['version']
        self.checksum = data.get("checksum", None)
        self.max_combo = data.get("max_combo", None)
        self.failtimes = Failtimes(data['failtimes']) if "failtimes" in data else None

        if 'beatmapset' in data and data['beatmapset'] is not None:
            if type(self).__name__ == 'Beatmap':
                self.beatmapset = Beatmapset(data['beatmapset'])
            else:
                self.beatmapset = BeatmapsetCompact(data['beatmapset'])
        else:
            self.beatmapset = None

    def __repr__(self):
        return prettify(self, 'version', 'beatmapset')


class BeatmapDifficultyAttributes:
    """
    Represent beatmap difficulty attributes. Following fields are always present and
    then there are additional fields for different rulesets.

    **Attributes**

    The parameters depend on the ruleset, but the following two attributes are present in all rulesets.

    max_combo: :class:`int`

    star_rating: :class:`float`

    mode_attributes: Union[:class:`OsuBeatmapDifficultyAttributes`, :class:`TaikoBeatmapDifficultyAttributes`,
    :class:`FruitsBeatmapDifficultyAttributes`, :class:`ManiaBeatmapDifficultyAttributes`, :class:`None`]
        Can be none for some beatmaps that are bugged and have no difficulty attributes.

    osu
        aim_difficulty: :class:`float`

        approach_rate: :class:`float`

        flashlight_difficulty: :class:`float`

        overall_difficulty: :class:`float`

        slider_factor: :class:`float`

        speed_difficulty: :class:`float`

    taiko
        stamina_difficulty: :class:`float`

        rhythm_difficulty: :class:`float`

        colour_difficulty: :class:`float`

        approach_rate: :class:`float`

        great_hit_window: :class:`float`

    fruits
        approach_rate: :class:`float`

    mania
        great_hit_window: :class:`float`

        score_multiplier: :class:`float`
    """
    __slots__ = (
        "max_combo", "star_rating", "type", "mode_attributes"
    )

    def __init__(self, data):
        data = data['attributes']
        self.max_combo = data['max_combo']
        self.star_rating = data['star_rating']
        if "aim_difficulty" in data:
            self.type = "osu"
            self.mode_attributes = OsuBeatmapDifficultyAttributes(data)
        elif "stamina_difficulty" in data:
            self.type = "taiko"
            self.mode_attributes = TaikoBeatmapDifficultyAttributes(data)
        elif "great_hit_window" in data:
            self.type = "mania"
            self.mode_attributes = ManiaBeatmapDifficultyAttributes(data)
        elif "approach_rate" in data:
            self.type = 'fruits'
            self.mode_attributes = FruitsBeatmapDifficultyAttributes(data)
        else:
            self.type = None
            self.mode_attributes = None

    def __getattr__(self, item):
        return getattr(self.mode_attributes, item)

    def __repr__(self):
        return prettify(self, 'star_rating', 'type', 'mode_attributes')


class OsuBeatmapDifficultyAttributes:
    """
    osu!standard beatmap difficulty attributes.
    See :class:`BeatmapDifficultyAttributes` for more information.

    **Attributes**

    aim_difficulty: :class:`float`

    approach_rate: :class:`float`

    flashlight_difficulty: :class:`float`

    overall_difficulty: :class:`float`

    slider_factor: :class:`float`

    speed_difficulty: :class:`float`

    speed_note_count: :class:`float`
    """
    __slots__ = (
        "aim_difficulty", "approach_rate", "flashlight_difficulty",
        "overall_difficulty", "slider_factor", "speed_difficulty",
        "speed_note_count"
    )

    def __init__(self, data):
        self.aim_difficulty = data['aim_difficulty']
        self.approach_rate = data['approach_rate']
        self.flashlight_difficulty = data['flashlight_difficulty']
        self.overall_difficulty = data['overall_difficulty']
        self.slider_factor = data['slider_factor']
        self.speed_difficulty = data['speed_difficulty']
        self.speed_note_count = data['speed_note_count']

    def __repr__(self):
        return prettify(self, 'aim_difficulty', 'speed_difficulty')


class TaikoBeatmapDifficultyAttributes:
    """
    osu!taiko beatmap difficulty attributes.
    See :class:`BeatmapDifficultyAttributes` for more information.

    **Attributes**

    stamina_difficulty: :class:`float`

    rhythm_difficulty: :class:`float`

    colour_difficulty: :class:`float`

    great_hit_window: :class:`float`

    peak_difficulty: :class:`float`
    """

    __slots__ = (
        "stamina_difficulty", "rhythm_difficulty", "colour_difficulty",
        "great_hit_window", "peak_difficulty"
    )

    def __init__(self, data):
        self.stamina_difficulty = data['stamina_difficulty']
        self.rhythm_difficulty = data['rhythm_difficulty']
        self.colour_difficulty = data['colour_difficulty']
        self.great_hit_window = data['great_hit_window']
        self.peak_difficulty = data['peak_difficulty']

    def __repr__(self):
        return prettify(self, 'stamina_difficulty')


class FruitsBeatmapDifficultyAttributes:
    """
    osu!catch beatmap difficulty attributes.
    See :class:`BeatmapDifficultyAttributes` for more information.

    **Attributes**

    approach_rate: :class:`float`
    """

    __slots__ = (
        "approach_rate"
    )

    def __init__(self, data):
        self.approach_rate = data['approach_rate']

    def __repr__(self):
        return prettify(self, 'approach_rate')


class ManiaBeatmapDifficultyAttributes:
    """
    osu!mania beatmap difficulty attributes.
    See :class:`BeatmapDifficultyAttributes` for more information.

    **Attributes**

    great_hit_window: :class:`float`
    """

    __slots__ = (
        "great_hit_window"
    )

    def __init__(self, data):
        self.great_hit_window = data['great_hit_window']

    def __repr__(self):
        return prettify(self, 'great_hit_window')


class Failtimes:
    """
    All attributes are optional but there's always at least one attribute present.

    **Attributes**

    exit: Sequence[:class:`int`]
        Sequence of integers. List is length 100.

    fail: Sequence[:class:`int`]
        Sequence of integers. List is length 100.
    """
    def __init__(self, data):
        if 'exit' in data:
            self.exit = data['exit']
        if 'fail' in data:
            self.fail = data['fail']

    def __repr__(self):
        try:
            return prettify(self, 'exit')
        except AttributeError:
            return prettify(self, 'fail')


class Beatmap(BeatmapCompact):
    """
    Represent a beatmap. This extends :class:`BeatmapCompact` with additional attributes.

    **Attributes**

    accuracy: :class:`float`

    ar: :class:`float`

    beatmapset_id: :class:`int`

    bpm: :class:`float` or :class:`NoneType`

    convert: :class:`bool`

    count_circles: :class:`int`

    count_sliders: :class:`int`

    count_spinners: :class:`int`

    cs: :class:`float`

    deleted_at: :class:`datetime.datetime` or :class:`NoneType`

    drain: :class:`float`

    hit_length: :class:`int`

    is_scoreable: :class:`bool`

    last_updated: :class:`datetime.datetime`

    mode_int: :class:`GameModeInt`

    passcount: :class:`int`

    playcount: :class:`int`

    ranked: :class:`RankStatus`

    url: :class:`str`
    """
    __slots__ = (
        "ranked", "url", "playcount", "passcount", "mode_int", "last_updated",
        "is_scoreable", "hit_length", "drain", "deleted_at", "cs", "count_spinners",
        "count_circles", "count_sliders", "convert", "bpm", "beatmapset_id", "ar",
        "accuracy"
    )

    def __init__(self, data):
        super().__init__(data)
        self.ranked = RankStatus(int(data['ranked']))
        self.url = data['url']
        self.playcount = data['playcount']
        self.passcount = data['passcount']
        self.mode_int = GameModeInt(data['mode_int'])
        self.last_updated = parser.parse(data['last_updated'])
        self.is_scoreable = data['is_scoreable']
        self.hit_length = data['hit_length']
        self.drain = data['drain']
        self.deleted_at = parser.parse(data['deleted_at']) if data['deleted_at'] is not None else None
        self.cs = data['cs']
        self.count_spinners = data['count_spinners']
        self.count_sliders = data['count_sliders']
        self.count_circles = data['count_circles']
        self.convert = data['convert']
        self.bpm = data['bpm']
        self.beatmapset_id = data['beatmapset_id']
        self.ar = data['ar']
        self.accuracy = data['accuracy']

    def __repr__(self):
        return super().__repr__()


class BeatmapPlaycount:
    """
    Represent the playcount of a beatmap.

    **Attributes**

    beatmap_id: :class:`int`

    beatmap: :class:`BeatmapCompact` or :class:`NoneType`

    beatmapset: :class:`BeatmapsetCompact` or :class:`NoneType`

    count: :class:`int`
    """
    __slots__ = (
        "beatmap_id", "beatmap", "beatmapset", "count"
    )

    def __init__(self, data):
        self.beatmap_id = data['beatmap_id']
        self.beatmap = BeatmapCompact(data['beatmap']) if data['beatmap'] is not None else None
        self.beatmapset = BeatmapsetCompact(data['beatmapset']) if data['beatmapset'] is not None else None
        self.count = data['count']

    def __repr__(self):
        return prettify(self, 'beatmap_id', 'count')



from dateutil import parser

from ..util import prettify
from ..enums import ForumTopicType


class ForumPost:
    """
    **Attributes**

    created_at: :class:`datetime.datetime`

    deleted_at: :class:`datetime.datetime` or :class:`NoneType`

    edited_at: :class:`datetime.datetime` or :class:`NoneType`

    edited_by_id: :class:`int` or :class:`NoneType`

    forum_id: :class:`int`

    id: :class:`int`

    topic_id: :class:`int`

    user_id: :class:`int`

    **Possible Attributes**

    body: :class:`dict`
        dictionary containing keys html and raw. html: Post content in HTML format. raw: content in BBCode format.
    """
    __slots__ = (
        "created_at", "deleted_at", "edited_at", "edited_by_id", "forum_id",
        "id", "topic_id", "user_id", "body"
    )

    def __init__(self, data):
        self.created_at = parser.parse(data["created_at"])
        self.deleted_at = parser.parse(data["deleted_at"]) if data["deleted_at"] is not None else None
        self.edited_at = parser.parse(data["edited_at"]) if data["edited_at"] is not None else None
        self.edited_by_id = data['edited_by_id']
        self.forum_id = data['forum_id']
        self.id = data['id']
        self.topic_id = data['topic_id']
        self.user_id = data['user_id']
        self.body = data.get('body', None)

    def __repr__(self):
        return prettify(self, 'user_id', 'topic_id')


class ForumTopic:
    """
    **Attributes**

    created_at: :class:`datetime.datetime`

    deleted_at: :class:`datetime.datetime` or :class:`NoneType`

    first_post_id: :class:`int`

    forum_id: :class:`int`

    id: :class:`int`

    is_locked: :class:`bool`

    last_post_id: :class:`int`

    poll: :class:`Poll` or :class:`NoneType`

    post_count: :class:`int`

    title: :class:`str`

    type: :class:`ForumTopicType`

    updated_at: :class:`datetime.datetime`

    user_id: :class:`int`
    """
    __slots__ = (
        "created_at", "deleted_at", "first_post_id", "forum_id", "id", "is_locked",
        "last_post_id", "poll", "post_count", "title", "type", "updated_at", "user_id"
    )

    def __init__(self, data):
        self.created_at = parser.parse(data["created_at"])
        self.deleted_at = parser.parse(data["deleted_at"]) if data["deleted_at"] is not None else None
        self.first_post_id = data['first_post_id']
        self.forum_id = data['forum_id']
        self.id = data['id']
        self.is_locked = data['is_locked']
        self.last_post_id = data['last_post_id']
        self.post_count = data['post_count']
        self.title = data['title']
        self.type = ForumTopicType(data['type'])
        self.updated_at = parser.parse(data["updated_at"])
        self.user_id = data['user_id']

    def __repr__(self):
        return prettify(self, 'user_id', 'title')


class Poll:
    """
    **Attributes**

    allow_vote_change: :class:`bool`

    ended_at: :class:`datetime.datetime` or :class:`NoneType`

    hide_incomplete_results: :class:`bool`

    last_vote_at: :class:`datetime.datetime` or :class:`NoneType`

    max_votes: :class:`int`

    options: Sequence[:class:`PollOption`]

    started_at: :class:`datetime.datetime`

    title: :class:`dict`[:class:`str`, :class:`str`]
        Has two items bbcode: :class:`str` and html: :class:`str`

    total_vote_count: :class:`int`
    """

    __slots__ = (
        "allow_vote_change", "ended_at", "hide_incomplete_results", "last_vote_at",
        "max_votes", "options", "started_at", "title", "total_vote_count"
    )

    def __init__(self, data):
        self.allow_vote_change = data['allow_vote_change']
        self.ended_at = parser.parse(data["ended_at"]) if data["ended_at"] is not None else None
        self.hide_incomplete_results = data['hide_incomplete_results']
        self.last_vote_at = parser.parse(data["last_vote_at"]) if data["last_vote_at"] is not None else None
        self.max_votes = data['max_votes']
        self.options = list(map(PollOption, data['options']))
        self.started_at = parser.parse(data["started_at"])
        self.title = data['title']
        self.total_vote_count = data['total_vote_count']

    def __repr__(self):
        return prettify(self, 'title', 'options')


class PollOption:
    """
    **Attributes**

    id: :class:`int`

    text: :class:`dict`[:class:`str`, :class:`str`]
        Has two items bbcode: :class:`str` and html: :class:`str`

    vote_count: :class:`int` or :class:`NoneType`
    """

    __slots__ = ("id", "text", "vote_count")

    def __init__(self, data):
        self.id = data['id']
        self.text = data['text']
        self.vote_count = data['vote_count']

    def __repr__(self):
        return prettify(self, 'text', 'vote_count')
